@@ -238,13 +238,16 @@ pum_display(
 	    // cmdline completion popup menu
 	    cursor_col = cmdline_compl_startcol();
 	else
+	{
+	    // w_wcol includes virtual text "above"
+	    int wcol = curwin->w_wcol % curwin->w_width;
 #ifdef FEAT_RIGHTLEFT
-	if (right_left)
-	    cursor_col = curwin->w_wincol + curwin->w_width
-							  - curwin->w_wcol - 1;
-	else
+	    if (right_left)
+		cursor_col = curwin->w_wincol + curwin->w_width - wcol - 1;
+	    else
 #endif
-	    cursor_col = curwin->w_wincol + curwin->w_wcol;
+		cursor_col = curwin->w_wincol + wcol;
+	}
 
 	// if there are more items than room we need a scrollbar
 	if (pum_height < size)
@@ -434,10 +437,17 @@ pum_redraw(void)
     int		round;
     int		n;
 
-    int *ha = highlight_attr;
-    //			 "word"		"kind"		"extra text"
-    int	attrsNorm[3] = { ha[HLF_PNI],	ha[HLF_PNK],	ha[HLF_PNX] };
-    int	attrsSel[3] =  { ha[HLF_PSI],	ha[HLF_PSK],	ha[HLF_PSX] };
+    int		attrsNorm[3];
+    int		attrsSel[3];
+    // "word"
+    attrsNorm[0] = highlight_attr[HLF_PNI];
+    attrsSel[0] = highlight_attr[HLF_PSI];
+    // "kind"
+    attrsNorm[1] = highlight_attr[HLF_PNK];
+    attrsSel[1] = highlight_attr[HLF_PSK];
+    // "extra text"
+    attrsNorm[2] = highlight_attr[HLF_PNX];
+    attrsSel[2] = highlight_attr[HLF_PSX];
 
     if (call_update_screen)
     {
